<?php
namespace Sharplogicians\Keycloak\Model;

use Magento\Framework\Controller\Result\JsonFactory;
use Magento\Customer\Api\CustomerRepositoryInterface;
use Magento\Framework\App\Config\ScopeConfigInterface;
use GuzzleHttp\Client as HttpClient;
use Magento\Framework\UrlInterface;
use Psr\Log\LoggerInterface;
use Magento\Framework\HTTP\Client\Curl;

class Keycloak
{
    protected $jsonFactory;

    protected $customerRepository;

    protected $scopeConfig;

    protected $httpClient;

    protected $urlBuilder;

    protected $curl;

    public function __construct(
        JsonFactory $jsonFactory,
        CustomerRepositoryInterface $customerRepository,
        ScopeConfigInterface $scopeConfig,
        \Magento\Store\Model\StoreManagerInterface $storeManager,
        \Magento\Customer\Model\CustomerFactory $customerFactory,
        UrlInterface $urlBuilder,
        Curl $curl
    ) {
        $this->jsonFactory = $jsonFactory;
        $this->customerRepository = $customerRepository;
        $this->scopeConfig = $scopeConfig;
        $this->httpClient = new HttpClient();
        $this->storeManager     = $storeManager;
        $this->customerFactory  = $customerFactory;
        $this->urlBuilder = $urlBuilder;
        $this->curl = $curl;
    }

    public function exchangeCodeForToken($code)
    {
        $clientId = $this->getClientId();
        $clientSecret = $this->getClientSecret();
        $baseUrl = $this->getKeyCloakBaseUrl();
        $realm = $this->getRealm();
        $redirectUri = $this->urlBuilder->getUrl('keycloak/login/callback');

        $tokenUrl = $baseUrl . '/realms/' . $realm . '/protocol/openid-connect/token';

        $response = $this->httpClient->post($tokenUrl, [
            'form_params' => [
                'grant_type' => 'authorization_code',
                'code' => $code,
                'redirect_uri' => $redirectUri,
                'client_id' => $clientId,
                'client_secret' => $clientSecret,
                'scope' => 'openid profile email'
            ],
            'verify' => false
        ]);

        return json_decode($response->getBody()->getContents(), true);
    }

    public function getUserInfo($accessToken)
    {
        $baseUrl = $this->getKeyCloakBaseUrl();
        $realm = $this->getRealm();
        $userInfoUrl = $baseUrl . '/realms/' . $realm . '/protocol/openid-connect/userinfo';

        $response = $this->httpClient->get($userInfoUrl, [
            'headers' => [
                'Authorization' => 'Bearer ' . $accessToken,
            ],
            'verify' => false
        ]);

        return json_decode($response->getBody()->getContents(), true);
    }

    public function getOrCreateCustomer($email, $name)
    {
        try {
            return $this->customerRepository->get($email);
        } catch (\Magento\Framework\Exception\NoSuchEntityException $e) {
            $fullname = explode(' ', $name);
            $firstName = $fullname[0];
            $lastName = $fullname[1];
            $customer = $this->createCustomer($email, $firstName, $lastName);
            return $customer;
        }
    }

    public function exportCustomer($customer, $accessToken, $password = '')
    {
        $keycloakBaseUrl = $this->getKeyCloakBaseUrl();
        $clientId = $this->getClientId();
        $clientSecret = $this->getClientSecret();
        $realm = $this->getRealm();

        $url = $keycloakBaseUrl . '/admin/realms/' . $realm . '/users';

        if($password) {
            $body = [
                'username' => $customer->getEmail(),
                'email' => $customer->getEmail(),
                'firstName' => $customer->getFirstname(),
                'lastName' => $customer->getLastname(),
                'enabled' => true,
                'credentials' => [
                    [
                        "type" => 'password',
                        "value" => $password,
                        "temporary" => false
                    ]
                ]
            ];
        } else {
            $body = [
                'username' => $customer->getEmail(),
                'email' => $customer->getEmail(),
                'firstName' => $customer->getFirstname(),
                'lastName' => $customer->getLastname(),
                'enabled' => true
            ];
        }        
        
        $response = $this->httpClient->post($url, [
            'headers' => [
                'Authorization' => 'Bearer ' . $accessToken,
                'Content-Type' => 'application/json'
            ],
            'json' => $body
        ]);

        return $response;
    }

    public function getAccessToken()
    {
        $keycloakBaseUrl = $this->getKeyCloakBaseUrl();
        $clientId = $this->getClientId();
        $clientSecret = $this->getClientSecret();
        $realm = $this->getRealm();

        $url = $keycloakBaseUrl . '/realms/' . $realm . '/protocol/openid-connect/token';

        try {
            $response = $this->httpClient->post($url, [
                'headers' => [
                    'Content-Type' => 'application/x-www-form-urlencoded', // Ensure the correct content type
                ],
                'form_params' => [
                    'grant_type' => 'client_credentials',
                    'client_id' => $clientId,
                    'client_secret' => $clientSecret
                ]
            ]);            

            $body = json_decode($response->getBody()->getContents(), true);
            return $body['access_token'] ?? null;
        } catch (\Exception $e) {
            return null;
        }
    }

    private function createCustomer($email, $firstName, $lastName)
    {
        try {
            $websiteId  = $this->storeManager->getWebsite()->getWebsiteId();
            $customer   = $this->customerFactory->create();
            $customer->setWebsiteId($websiteId);
            $customer->setEmail($email);
            $customer->setFirstname($firstName);
            $customer->setLastname($lastName);
            $customer->setPassword($this->generateRandomPassword());
            $customer->save();
            $customer->setGroupId(1);
            return $this->customerRepository->getById($customer->getId());
        } catch (\Exception $e) {
            throw new \Exception('Unable to create customer: ' . $e->getMessage());
        }
    }

    public function getUserIdByEmail($email, $accessToken)
    {
        $keycloakBaseUrl = $this->getKeyCloakBaseUrl();
        $realm = $this->getRealm();

        if (!$accessToken) {
            return null;
        }

        $url = $keycloakBaseUrl . "/admin/realms/" . $realm . "/users?email=" . urlencode($email);

        $this->curl->setHeaders([
            'Authorization' => 'Bearer ' . $accessToken,
            'Accept' => 'application/json'
        ]);

        $this->curl->get($url);
        $response = json_decode($this->curl->getBody(), true);

        if (!empty($response) && isset($response[0]['id'])) {
            return $response[0]['id'];
        }

        return null;
    }

    public function updateCustomer($email, $accessToken, $newPassword)
    {
        $keycloakBaseUrl = $this->getKeyCloakBaseUrl();
        $realm = $this->getRealm();

        $userId = $this->getUserIdByEmail($email, $accessToken);
        if (!$userId) {
            return ['success' => false, 'message' => "User not found for email: " . $email];
        }
        
        if (!$accessToken) {
            return ['success' => false, 'message' => "Failed to obtain Keycloak admin access token."];
        }

        $url = $keycloakBaseUrl . "/admin/realms/" . $realm . "/users/" . $userId . "/reset-password";

        $payload = json_encode([
            "type" => "password",
            "value" => $newPassword,
            "temporary" => false
        ]);

        $this->curl->setHeaders([
            'Authorization' => 'Bearer ' . $accessToken,
            'Content-Type' => 'application/json'
        ]);

        $this->curl->setOption(CURLOPT_CUSTOMREQUEST, 'PUT');
        $this->curl->post($url, $payload);

        if ($this->curl->getStatus() == 204) {
            return ['success' => true, 'message' => "Password updated successfully."];
        } else {
            return ['success' => false, 'message' => "Failed to update password."];
        }
    }

    public function authenticateUser($username, $password)
    {
        $keycloakBaseUrl = $this->getKeyCloakBaseUrl();
        $clientId = $this->getClientId();
        $clientSecret = $this->getClientSecret();
        $realm = $this->getRealm();

        $url = $keycloakBaseUrl . "/realms/" . $realm . "/protocol/openid-connect/token";

        $params = [
            'client_id' => $clientId,
            'client_secret' => $clientSecret,
            'username' => $username,
            'password' => $password,
            'grant_type' => 'password'
        ];

        $this->curl->setOption(CURLOPT_RETURNTRANSFER, true);
        $this->curl->post($url, $params);

        $response = json_decode($this->curl->getBody(), true);

        if (isset($response['access_token'])) {
            return $response['access_token'];
        } else {
            return false;
        }
    }

    public function logoutUser($accessToken)
    {
        $keycloakBaseUrl = $this->getKeyCloakBaseUrl();
        $clientId = $this->getClientId();
        $clientSecret = $this->getClientSecret();
        $realm = $this->getRealm();

        $logoutUrl = $keycloakBaseUrl . "/realms/" . $realm . "/protocol/openid-connect/logout";

        // Prepare logout request
        $params = [
            'client_id' => $clientId,
            'client_secret' => $clientSecret,
            'refresh_token' => $accessToken
        ];

        $this->curl->setHeaders([
            'Content-Type' => 'application/x-www-form-urlencoded'
        ]);

        $this->curl->post($logoutUrl, http_build_query($params));

        return $this->curl->getBody();
    }

    private function generateRandomPassword($length = 12)
    {
        return bin2hex(random_bytes($length / 2));
    }

    private function getKeyCloakBaseUrl()
    {
        return $this->scopeConfig->getValue('keycloak_sso/general/keycloak_base_url', \Magento\Store\Model\ScopeInterface::SCOPE_STORE);
    }

    private function getClientId()
    {
        return $this->scopeConfig->getValue('keycloak_sso/general/client_id', \Magento\Store\Model\ScopeInterface::SCOPE_STORE);
    }

    private function getClientSecret()
    {
        return $this->scopeConfig->getValue('keycloak_sso/general/client_secret', \Magento\Store\Model\ScopeInterface::SCOPE_STORE);
    }

    private function getRealm()
    {
        return $this->scopeConfig->getValue('keycloak_sso/general/realm', \Magento\Store\Model\ScopeInterface::SCOPE_STORE);
    }
}
